/*
LockDis.  A very simple X screen locker.  No flying toasters.
Copyright (C) 2009  Brett McNamara <brett@chaingang.org>

This file is part of LockDis.

LockDis is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Foobar is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>

#include "lockdis.h"

#include "img/mask_char.xpm"
#include "img/logo.xpm"


static Window getScreenWindow(Display *display, int screenNum) {
	XSetWindowAttributes attr;
	
	attr.override_redirect = True;
	attr.background_pixel = attr.border_pixel = BlackPixel(display, screenNum);
	attr.event_mask = ExposureMask | VisibilityChangeMask | StructureNotifyMask;
	attr.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
	return XCreateWindow(display, RootWindow(display, screenNum), 
		0, 0, DisplayWidth(display, screenNum), DisplayHeight(display, screenNum), 
		0, CopyFromParent, InputOutput, CopyFromParent,
		CWOverrideRedirect | CWBackPixel | CWBorderPixel | CWEventMask | CWDontPropagate, 
		&attr
	);
}


static void initScreenGc(DisplayInfo *di) {
	XGCValues gv;
	unsigned long black = BlackPixel(di->display, di->defaultScreenNum);

	gv.foreground = gv.background = black;
	di->gc = XCreateGC(di->display, *(di->windows), GCForeground | GCBackground, &gv);
	XSetForeground(di->display, di->gc, black);  // why do I have to do this?  Who knows.
}


static Cursor getBlankCursor(Display *dpy, Window win) {
	static char data[1] = {0};
	XColor dummy;
	Pixmap blank = XCreateBitmapFromData (dpy, win, data, 1, 1);
	Cursor cursor = XCreatePixmapCursor(dpy, blank, blank, &dummy, &dummy, 0, 0);

	XFreePixmap (dpy, blank);
	
	return cursor;
}


static void initScreenHideCursor(Display *display, Window win) {
	XDefineCursor(display, win, getBlankCursor(display, win));
}


static void initScreens(DisplayInfo *di) {
	int i, winNum;
	
	di->windowsCount = ScreenCount(di->display);
	di->defaultScreenNum = DefaultScreen(di->display);
	
	di->windows = (Window *) malloc(sizeof(Window) * di->windowsCount);

	*di->windows = getScreenWindow(di->display, di->defaultScreenNum);
	di->width = DisplayWidth(di->display, di->defaultScreenNum);
	di->height = DisplayHeight(di->display, di->defaultScreenNum);
	initScreenGc(di);
	initScreenHideCursor(di->display, *di->windows);
	XMapRaised(di->display, *di->windows);
	
	winNum=1;
	for (i = 0; i < di->windowsCount; i++) {
		if (i!=di->defaultScreenNum) {
			Window *pWin = &di->windows[winNum++];
			*pWin = getScreenWindow(di->display, di->defaultScreenNum);
			initScreenHideCursor(di->display, *pWin);
			XMapRaised(di->display, *pWin);
		}
	}
}


static XImage *getImage(DisplayInfo *di, char *xpm[]) {
	XImage *img;
	
	if (XpmCreateImageFromData (di->display, xpm, &img, NULL, NULL)) {
		fprintf(stderr, "Can't read image.\n");
		exit(EXIT_FAILURE);
	}
	return img;
}


static XImage *getLogoImage(DisplayInfo *di) {
	static XImage *img = NULL;
	if (img==NULL) { img = getImage(di, logo_xpm); }
	return img;
}


static XImage *getMaskImage(DisplayInfo *di) {
	static XImage *img = NULL;
	if (img==NULL) { img = getImage(di, mask_char_xpm); }
	return img;
}


static void putLogo(DisplayInfo *di, int size) {
	XImage *img = getLogoImage(di);
	int x = (di->width - img->width)/2;
	int y = (di->height - img->height)/4;
	
	XPutImage(di->display, *di->windows, di->gc, img, 0, 0, x, y, img->width, img->height );
}


static void putEntry(DisplayInfo *di, int size) {
	XImage *img = getMaskImage(di);
	int y = di->height/2 - img->height/2;
	int x = (di->width - (img->width*size))/2;
	int maxSize = (((di->width - img->width)/2) / img->width)*2;
	
	if (size>maxSize) {
		//int odd = (size%2!=0);
		x = (di->width - (img->width*maxSize))/2;
		if (size%2!=0) {
			x -= img->width/2;
			size = maxSize + 1;
		} else {
			size = maxSize;
		}
	}
	
	XFillRectangle(di->display, *di->windows, di->gc, 0, y, di->width, img->height);
	while(size-- > 0) {
		XPutImage(di->display, *di->windows, di->gc, img, 0, 0, x, y, img->width, img->height );
		x += img->width;
	}
}


void showScreen(DisplayInfo *di, int size) {
	if (size==0) {
		XClearWindow(di->display, *di->windows);
	} else {
		// should work, sometimes refresh wonky
		// if (size==1) { putLogo(di, size); } 
		putLogo(di, size);
		putEntry(di, size);
	}
}


void raiseWindows(DisplayInfo *di) {
	int i;

	for (i=0; i<di->windowsCount; i++) {
		XRaiseWindow(di->display, di->windows[i]);
	}
}


void initDisplay(DisplayInfo *di) {
	di->display = XOpenDisplay("");
	if (di->display == 0) { 
		fprintf(stderr, "Can't open X.\n");
		exit(EXIT_FAILURE);
	}
	
	signal(SIGTERM, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	
	initScreens(di);
	
	XGrabKeyboard(di->display, *di->windows, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	XSync(di->display, False);
}
